﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Timers;
using System.Xml;
using System.Globalization;

namespace HabboTraxDEBESTE2
{
    public class TraxPage
    {
        private TraxPart[,] traxOnPage;
        private Timer playTimer;

        public TraxPage()
        {
            traxOnPage = new TraxPart[24, 4];
        }

        public int Length
        {
            get { return traxOnPage.GetLength(0); }
        }
        public int Height
        {
            get { return traxOnPage.GetLength(1); }
        }

        public int FinalTraxTime
        {
            get
            {
                int finalPos = 0;
                for (int i = 0; i < traxOnPage.GetLength(0); i++)
                    for (int j = 0; j < traxOnPage.GetLength(1); j++)
                        if (traxOnPage[i, j] != null && i > finalPos)
                            finalPos = i;
                return finalPos;
            }
        }

        public void AddSlots(int numberOfSlots)
        {
            TraxPart[,] newPageTrax = new TraxPart[traxOnPage.GetLength(0) + numberOfSlots, traxOnPage.GetLength(1)];
            for (int i = 0; i < numberOfSlots; i++)
                for (int j = 0; j < traxOnPage.GetLength(1); j++)
                    if (traxOnPage[i, j] != null)
                        newPageTrax[i, j] = traxOnPage[i, j];
            traxOnPage = newPageTrax;
        }

        public bool HasNextPart(int time, int row)
        {
            return traxOnPage[time, row].part < traxOnPage[time, row].TotalParts - 1;
        }

        /// <summary>
        /// Adds the given Trax to the TraxPage at the provided position. If overwrite is true, other trax 
        /// that occupy the same tile will be removed from the page
        /// </summary>
        /// <param name="trax"></param>
        /// <param name="time"></param>
        /// <param name="row"></param>
        /// <param name="overwrite"></param>
        public void AddTrax(Trax trax, int time, int row, bool overwrite)
        {
            int nrOfParts = trax.GetTotalParts();
            if (!overwrite && nrOfParts > 1)
            { // Check for trax occupying the same position
                for (int i = 0; i < nrOfParts; i++)
                    if (traxOnPage[time + 1, row] != null)
                        throw new TileOccupiedException(row, time + 1);
            }
            TraxPart previousPart = null;
            for (int i = 0; i < nrOfParts; i++)
            {
                TraxPart toAdd = new TraxPart(trax, i);
                if (traxOnPage[time + i, row] != null)
                {
                    if (overwrite)
                        RemoveTrax(time + i, row);
                    else
                        throw new TileOccupiedException(row, time + 1);
                }
                traxOnPage[time + i, row] = new TraxPart(trax, i);
                previousPart = toAdd;
            }
        }

        public void RemoveTrax(int time, int row)
        {
            if (traxOnPage[time, row] == null)
                return;
            TraxPart toDel = traxOnPage[time, row];
            traxOnPage[time, row] = null;

            if (time > 0)
            {
                TraxPart prev = traxOnPage[time - 1, row];
                if (prev != null && prev.Trax == toDel.Trax && prev.part == toDel.part - 1 && prev.TotalParts == toDel.TotalParts)
                    RemoveTrax(time - 1, row);
            }
            if (time < traxOnPage.GetLength(0))
            {
                TraxPart next = traxOnPage[time + 1, row];
                if (next != null && next.Trax == toDel.Trax && next.part == toDel.part + 1 && next.TotalParts == toDel.TotalParts)
                    RemoveTrax(time + 1, row);
            }

        }



        public Trax GetTrax(int time, int row)
        {
            if (traxOnPage[time, row] == null)
                return null;
            return traxOnPage[time, row].Trax;
        }

        private int slotsPlayed = 0;

        public void Play(Player player, bool nextPageExists)
        {
            playTimer = new Timer(2000);
            playTimer.AutoReset = true;
            slotsPlayed = 0;
            playTimer.Elapsed += (s, e) => 
                {
                    if (nextPageExists && slotsPlayed >= traxOnPage.GetLength(0) || !nextPageExists && slotsPlayed > FinalTraxTime)
                    {
                        playTimer.Stop();
                        playTimer.Dispose(); //Making misuse of the dispose event to see when we finished playying.
                        slotsPlayed = 0;
                        return;
                    }
                    PlayTimeSlot(slotsPlayed, player);
                    slotsPlayed++;
                };
            playTimer.Start();
            playTimer.Disposed += FinishedPlaying;
            PlayTimeSlot(slotsPlayed, player);
            slotsPlayed++;
        }

        private void PlayTimeSlot(int time, Player player)
        {
            Console.WriteLine("Playing slot " + slotsPlayed);
            List<Trax> TraxToPlayNow = new List<Trax>();
            for (int j = 0; j < traxOnPage.GetLength(1); j++)
                if (traxOnPage[slotsPlayed, j] != null && traxOnPage[slotsPlayed, j].part == 0) //First part of the trax
                    TraxToPlayNow.Add(traxOnPage[slotsPlayed, j].Trax);
            TraxToPlayNow.ForEach(tr => player.PlaySound(tr));
            if (OnSlotStartedPlaying != null)
                OnSlotStartedPlaying(this, new SlotEventArgs() { Slot = time });
        }

        public EventHandler FinishedPlaying;
        public EventHandler<SlotEventArgs> OnSlotStartedPlaying;

        public void Stop()
        {
            playTimer.Stop();
            playTimer.Dispose();
        }

        public void LoadTraxToPlayer(Player player)
        {
            List<Trax> traxNeeded = new List<Trax>();
            for(int i = 0; i < traxOnPage.GetLength(0); i++)
                for (int j = 0; j < traxOnPage.GetLength(1); j++)
                    if (traxOnPage[i, j] != null && traxOnPage[i, j].Trax != null && !traxNeeded.Contains(traxOnPage[i,j].Trax))
                        traxNeeded.Add(traxOnPage[i,j].Trax);

            player.TraxNeeded = traxNeeded.ToArray();
        }

        public void WriteToXml(XmlWriter textWriter)
        {
            textWriter.WriteStartElement("Page");
            textWriter.WriteAttributeString("length", traxOnPage.GetLength(0).ToString(CultureInfo.InvariantCulture));
            textWriter.WriteAttributeString("rows", traxOnPage.GetLength(1).ToString(CultureInfo.InvariantCulture));
            for (int i = 0; i < traxOnPage.GetLength(0); i++)
            {
                textWriter.WriteStartElement("TimeSlot");
                textWriter.WriteAttributeString("number", i.ToString(CultureInfo.InvariantCulture));
                for (int j = 0; j < traxOnPage.GetLength(1); j++)
                {
                    if (traxOnPage[i, j] != null && traxOnPage[i,j].Trax != null)
                    {
                        textWriter.WriteStartElement("TraxPart");
                        textWriter.WriteAttributeString("traxCD", traxOnPage[i, j].Trax.CD);
                        textWriter.WriteAttributeString("traxNr", traxOnPage[i, j].Trax.TraxNr.ToString(CultureInfo.InvariantCulture));
                        textWriter.WriteAttributeString("partNr", traxOnPage[i, j].part.ToString(CultureInfo.InvariantCulture));
                        //textWriter.WriteAttributeString("NrOfParts", traxOnPage[i, j].TotalParts.ToString(CultureInfo.InvariantCulture));
                        textWriter.WriteAttributeString("row", j.ToString(CultureInfo.InvariantCulture));
                        textWriter.WriteEndElement();
                    }
                }
                textWriter.WriteEndElement();
            }
            textWriter.WriteEndElement();
        }

        public static TraxPage LoadFromXml(XmlReader reader, out Trax[] TraxToLoadOut)
        {
            int length = int.Parse(reader.GetAttribute("length"));
            int rows = int.Parse(reader.GetAttribute("rows"));
            TraxPage newPage = new TraxPage();
            newPage.traxOnPage = new TraxPart[length, rows];
            //List<Trax> traxToLoad = new List<Trax>();
            int currentTimeSlot = -1;
            Dictionary<string, Trax> usedTrax = new Dictionary<string,Trax>(); //Keep a list of used trax to save memory (no new Trax object for each slot)
            while (reader.Read() && !(reader.NodeType == XmlNodeType.EndElement && reader.Name == "Page")) //Read until the end of the traxpage
            {
                if (reader.NodeType == XmlNodeType.Element && reader.Name == "TimeSlot")
                    try
                    {
                        currentTimeSlot = int.Parse(reader.GetAttribute("number"));
                    }
                    catch
                    {
                        currentTimeSlot++;
                    }

                if (reader.NodeType == XmlNodeType.Element && reader.Name == "TraxPart")
                {
                    int partNr = int.Parse(reader.GetAttribute("partNr"));
                    int row = int.Parse(reader.GetAttribute("row"));
                    Trax traxToAdd = new Trax(reader.GetAttribute("traxCD"), int.Parse(reader.GetAttribute("traxNr")));
                    if (usedTrax.ContainsKey(traxToAdd.Code))
                        traxToAdd = usedTrax[traxToAdd.Code];
                    else
                        usedTrax.Add(traxToAdd.Code, traxToAdd);
                    newPage.traxOnPage[currentTimeSlot, row] = new TraxPart(traxToAdd, partNr);
                }

            }

            TraxToLoadOut = usedTrax.Values.ToArray();
            return newPage;
        }
    }
}
